from __future__ import annotations

from typing import Any, NamedTuple, Union

import torch
from torch import nn
from torchview import FunctionNode, ModuleNode, TensorNode, draw_graph

from rebasin._paths import ModelPaths
from rebasin.structs import (
    ModuleParameters,
    Permutation,
)

NODE_TYPES = Union[ModuleNode, FunctionNode, TensorNode]


class PermInfo(NamedTuple):
    module_parameters: ModuleParameters
    axis: int


class PermutationInitialization:
    def __init__(
            self,
            model_a: nn.Module,
            model_b: nn.Module,
            input_data_b: Any,
            input_data_a: Any | None = None
    ) -> None:
        self.model_a = model_a
        self.model_b = model_b
        self._id_to_module_a: dict[int, nn.Module] = {
            id(m): m for m in model_a.modules()
        }
        self._id_to_module_b: dict[int, nn.Module] = {
            id(m): m for m in model_b.modules()
        }

        self.input_data_b = input_data_b
        self.input_data_a = input_data_a if input_data_a is not None else input_data_b
        self.paths = self._trace_path()
        self.paths.merge_permutations()

        # For initializing the permutations, it makes most sense to track the modules,
        #   but for optimizing the permutations, it is more natural to go through the
        #   permutations. Therefore, I now switch to a datastructure fit for
        #   that purpose.
        self.perm_to_info: list[
            tuple[Permutation, list[tuple[int, ModuleParameters]]]
        ] = []
        self._get_permutations()

    def _trace_path(self) -> ModelPaths:
        root_a = list(
            draw_graph(self.model_a, input_data=self.input_data_a, depth=1e12)
            .root_container
        )
        root_b = list(
            draw_graph(self.model_b, input_data=self.input_data_b, depth=1e12)
            .root_container
        )

        if len(root_a) != len(root_b):
            raise ValueError(
                "Both models must have the same architecture!"
            )

        visited_nodes: set[NODE_TYPES] = set()
        working_nodes_a: list[NODE_TYPES] = root_a  # type: ignore[assignment]
        working_nodes_b: list[NODE_TYPES] = root_b  # type: ignore[assignment]
        paths: list[list[ModuleParameters]] = []

        while working_nodes_a and working_nodes_b:
            assert len(working_nodes_a) == len(working_nodes_b), \
                "Both models must have the same architecture!"

            new_working_nodes_a: list[NODE_TYPES] = []
            new_working_nodes_b: list[NODE_TYPES] = []

            for node_a, node_b in zip(working_nodes_a, working_nodes_b):  # noqa: B905
                if node_a in visited_nodes or node_b in visited_nodes:
                    continue
                path, children_a, children_b, visited = self._trace_path_from_node(
                    node_a, node_b
                )
                paths.append(path)
                new_working_nodes_a.extend(children_a)
                new_working_nodes_b.extend(children_b)
                visited_nodes.update(visited)

            working_nodes_a = new_working_nodes_a
            working_nodes_b = new_working_nodes_b

        return ModelPaths(paths)

    def _trace_path_from_node(
            self,
            node_a: NODE_TYPES,
            node_b: NODE_TYPES
    ) -> tuple[
        list[ModuleParameters],
        list[NODE_TYPES],
        list[NODE_TYPES],
        set[NODE_TYPES]
    ]:
        # Steps:
        # 1. Trace node_b until there are no children or a child has two parents
        #     or the node has multiple children
        # 2. Add the children to the children_b
        # 3. Do the same for node_a.
        # 4. In node_b, add ModuleParameters to the path when possible
        # 5. Make sure that both nodes always have the same children
        # 6. Trace visited nodes
        path: list[ModuleParameters] = []
        visited: set[NODE_TYPES] = set()

        while (
            len(node_b.children) == 1
            and len(tuple(node_b.children)[0].parents) == 1
        ):
            assert len(node_a.children) == 1, \
                "Both models must have the same architecture!"
            assert len(tuple(node_a.children)[0].parents) == 1, \
                "Both models must have the same architecture!"

            self._extend_path(path, node_a, node_b)
            visited.update([node_a, node_b])
            node_a = tuple(node_a.children)[0]  # type: ignore[assignment]
            node_b = tuple(node_b.children)[0]  # type: ignore[assignment]

        children_a: list[NODE_TYPES] = list(node_a.children)  # type: ignore[arg-type]
        children_b: list[NODE_TYPES] = list(node_b.children)  # type: ignore[arg-type]

        return path, children_a, children_b, visited

    def _extend_path(
            self,
            path: list[ModuleParameters],
            node_a: NODE_TYPES,
            node_b: NODE_TYPES
    ) -> None:
        assert type(node_a) == type(node_b), \
            "Both models must have the same architecture! " \
            f"But {type(node_a)=} and {type(node_b)=}"

        if not isinstance(node_a, ModuleNode) or not isinstance(node_b, ModuleNode):
            return

        module_a = self._id_to_module_a[node_a.compute_unit_id]
        module_b = self._id_to_module_b[node_b.compute_unit_id]

        # Only use leaf nodes
        if list(module_a.children()):
            assert module_b.children(), \
                "Both models must have the same architecture!"
            return

        w_a: nn.Parameter | None = None
        w_b: nn.Parameter | None = None
        b_a: nn.Parameter | None = None
        b_b: nn.Parameter | None = None

        for (name_a, param_a), (name_b, param_b) in zip(  # noqa: B905
                module_a.named_parameters(),
                module_b.named_parameters(),
        ):
            assert name_a == name_b, \
                "Both models must have the same architecture!"

            if "weight" in name_a:
                w_a = param_a
                w_b = param_b
            elif "bias" in name_a:
                b_a = param_a
                b_b = param_b

        if w_a is not None:
            assert w_b is not None, \
                "Both models must have the same architecture!"

            axis_to_permutation = self._get_axis_to_permutation(w_a, w_b)
            path.append(
                ModuleParameters(
                    weight_a=w_a,
                    weight_b=w_b,
                    name=module_b.__class__.__name__,
                    axis_to_permutation=axis_to_permutation,
                    module_type=type(module_b),
                    bias_a=b_a,
                    bias_b=b_b
                )
            )
        if w_b is not None:
            assert w_a is not None, \
                "Both models must have the same architecture!"

    @staticmethod
    def _get_axis_to_permutation(
            w_a: nn.Parameter, w_b: nn.Parameter
    ) -> dict[int, Permutation]:
        """Get the axis to permutation mapping."""
        assert w_a.shape == w_b.shape, \
            "Both models must have the same architecture!"

        axis_to_permutation = {
            0: Permutation(torch.arange(w_a.shape[0]))
        }
        if len(w_a.shape) > 1:
            axis_to_permutation[1] = Permutation(torch.arange(w_a.shape[1]))

        return axis_to_permutation

    def _get_permutations(self) -> None:
        """Get the permutations."""
        id_to_info: dict[
            int,
            tuple[Permutation, list[tuple[int, ModuleParameters]]]
        ] = {}

        for path in self.paths.paths:
            for module_params in path:
                for axis, perm in module_params.axis_to_permutation.items():
                    if id(perm) in id_to_info:
                        id_to_info[id(perm)][1].append((axis, module_params))
                    else:
                        id_to_info[id(perm)] = (perm, [(axis, module_params)])

        for info in id_to_info.values():
            self.perm_to_info.append(info)



"""Tests for the PermutationCoordinateDescent permutation initialization,
as defined in rebasin/initialization/_initializer.py: PermutationInitialization.
"""

from __future__ import annotations

import copy

import torch
from torch import nn
from torchvision.models import resnet50  # type: ignore[import]

from rebasin._initializer import PermutationInitialization
from tests.fixtures.models import MLP
from tests.fixtures.utils import (
    allclose,
    model_change_percent,
    randomize_permutations,
    reset_bn_running_stats,
)


def test_on_mlp() -> None:
    """Test the :class:`PermutationInitialization` initializer on an MLP."""
    common_tests(MLP(25, 10), MLP(25, 10), torch.randn(25))


def test_on_resnet50() -> None:
    """Test the :class:`PermutationInitialization` initializer on ResNet50."""
    common_tests(resnet50(), resnet50(), torch.randn(1, 3, 224, 224))


def common_tests(model_a: nn.Module, model_b: nn.Module, x: torch.Tensor) -> None:
    model_b_orig = copy.deepcopy(model_b)
    reset_bn_running_stats(model_b)
    y_orig = model_b(x)

    initializer = PermutationInitialization(model_a, model_b, x)
    randomize_permutations(initializer)
    initializer.paths.apply_permutations()
    assert model_change_percent(model_b, model_b_orig) > 0.1

    reset_bn_running_stats(model_b)
    y_new = model_b(x)
    assert allclose(y_orig, y_new)



from __future__ import annotations

import copy
import sys
from typing import Any

import pytest
import torch
from torch import nn
from torchvision.models import resnet18  # type: ignore[import]

from rebasin import PermutationCoordinateDescent
from rebasin.permutation_coordinate_descent import calculate_progress
from tests.fixtures.models import MLP
from tests.fixtures.utils import model_change_percent, model_distance, model_similarity


def test_calculate_progress() -> None:
    """Test the calculation of the progress."""
    cost_mat = torch.tensor([[0.0, 1.0, 2.0], [1.0, 0.0, 3.0], [2.0, 3.0, 0.0]])
    perm_old = torch.tensor([0, 1, 2])
    perm_new = torch.tensor([2, 0, 1])

    progress = calculate_progress(cost_mat, perm_old, perm_new)

    assert progress

    perm_old = torch.tensor([2, 0, 1])
    perm_new = torch.tensor([0, 1, 2])

    progress = calculate_progress(cost_mat, perm_old, perm_new)

    assert not progress


class TestPermutationCoordinateDescent:
    @pytest.mark.skipif("--full-suite" not in sys.argv, reason="Slow test")
    def test_resnet18(self) -> None:
        self.common_tests(torch.randn(1, 3, 224, 224), resnet18)

    def test_mlp(self) -> None:
        in_features, num_layers = 50, 5
        self.common_tests(torch.randn(50), MLP, in_features, num_layers)

    # TODO: Fix this test
    @pytest.mark.xfail(reason="Currently has problem, fix later")
    def test_multihead_attention(self) -> None:
        embed_dim = num_heads = 32
        x = torch.randn(embed_dim, num_heads)
        self.common_tests((x, x, x), nn.MultiheadAttention, embed_dim, num_heads)

    @staticmethod
    def common_tests(
            input_data: Any, constructor: Any, *args: Any
    ) -> None:
        for _ in range(10):
            model_a = constructor(*args)
            model_b = constructor(*args)

            model_b_old = copy.deepcopy(model_b)  # for comparison
            pcd = PermutationCoordinateDescent(model_a, model_b, input_data)
            pcd.calculate_permutations()
            pcd.apply_permutations()

            assert (
                    model_distance(model_a, model_b)
                    < model_distance(model_a, model_b_old)
            )


@pytest.mark.skipif(not torch.cuda.is_available(), reason="GPU test")
class TestPCDOnGPU:
    @staticmethod
    def test_mlp() -> None:
        device_b = torch.device("cuda")
        model_a, model_b = MLP(25), MLP(25).to(device_b)
        model_b_old = copy.deepcopy(model_b)

        pcd = PermutationCoordinateDescent(
            model_a, model_b, torch.randn(25), device_a="cpu", device_b=device_b
        )
        pcd.calculate_permutations()
        pcd.apply_permutations()

        assert (
                model_similarity(model_a, model_b)
                > model_similarity(model_a, model_b_old)
        )



"""
The path through the model as parametrized by ~ModuleParameters.
"""

from __future__ import annotations

from collections.abc import Sequence

import torch
from torch import nn

from rebasin import utils
from rebasin.structs import ModuleParameters, Permutation


def merge_linear_path(path: Sequence[ModuleParameters]) -> None:
    if len(path) < 2:
        return

    for mod0, mod1 in utils.pairwise(path):
        # for mypy:
        assert isinstance(mod0, ModuleParameters)
        assert isinstance(mod1, ModuleParameters)

        # Make the ModuleParameter.output_permutation.setter fix any issues.
        # This is important because nD-LayerNorms with n>1 must not have
        #   their output dimension changed.
        mod0.output_permutation = mod0.output_permutation
        mod1.output_permutation = mod1.output_permutation

        # LayerNorms behave differently than other modules
        #   (see test_output_consistency).
        if (
            mod0.module_type is nn.LayerNorm
            and len(mod0.input_permutation) == len(mod1.input_permutation)
        ):
            mod1.input_permutation = mod0.input_permutation
        elif len(mod0.output_permutation) == len(mod1.input_permutation):
            mod1.input_permutation = mod0.output_permutation


class ModelPaths:
    def __init__(self, paths: Sequence[Sequence[ModuleParameters]]) -> None:
        self.paths = paths

    def apply_permutations(self) -> None:
        # Make sure that the constraints on the permutations are applied:
        #  1. They are merged
        #  2. They change the model in such a way
        #       that it produces the same output as before
        self.merge_permutations()

        # Apply the permutations to the model
        for path in self.paths:
            for mod in path:
                mod.apply_permutations()

    def merge_permutations(self) -> None:
        for path in self.paths:
            if not path:
                continue
            self._make_input_identity(path)
            merge_linear_path(path)
            self._make_output_identity(path)

    @staticmethod
    def _make_input_identity(path: Sequence[ModuleParameters]) -> None:
        """
        Make the input permutation of the first module in the path the identity.

        For one-dimensional weights, the input permutation is also the output
        permutation. Therefore, we need to make
        """
        in_pt = 0  # The input permutation pointer
        identity = Permutation(torch.arange(len(path[in_pt].input_permutation)))
        path[in_pt].input_permutation = identity

        # Handle multi-dim LayerNorm inputs
        path[in_pt].output_permutation = path[in_pt].output_permutation

        # Handle 1d LayerNorm inputs
        while len(path[in_pt].axis_to_permutation) == 1 and in_pt < len(path) - 1:
            in_pt += 1
            path[in_pt].input_permutation = path[in_pt - 1].output_permutation

    @staticmethod
    def _make_output_identity(path: Sequence[ModuleParameters]) -> None:
        """
        Make the output permutation of the last module in the path the identity.

        For one-dimensional weights, the input permutation is also the output
        permutation. Therefore, we need to make
        """
        out_pt = -1  # The output permutation pointer
        identity = Permutation(torch.arange(len(path[out_pt].output_permutation)))
        path[out_pt].output_permutation = identity

        # Handle LayerNorm as well as 1d outputs
        while (
                (
                        path[out_pt].module_type is nn.LayerNorm
                        or len(path[out_pt].axis_to_permutation) == 1
                )
                and -out_pt < len(path)
        ):
            out_pt -= 1
            identity = Permutation(
                torch.arange(len(path[out_pt + 1].input_permutation))
            )
            path[out_pt + 1].input_permutation = identity
            path[out_pt].output_permutation = path[out_pt + 1].input_permutation



from __future__ import annotations

from typing import Any

import pytest
import torch
from torch import nn
from torch.utils.data import DataLoader

from rebasin import interpolation as interp
from tests.fixtures.datasets import RandomDataset
from tests.fixtures.models import MLP


def loss_fn(x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    return ((x - y) ** 2).sum()  # type: ignore[no-any-return]


def test_loss_fn() -> None:
    """Test that loss_fn is working as expected."""
    l1 = loss_fn(torch.tensor([1.0]), torch.tensor([2.0]))
    l2 = loss_fn(torch.tensor([1.0]), torch.tensor([1.0]))
    assert l1 > l2


class BaseClass:
    @staticmethod
    def loss_fn(x: torch.Tensor, y: torch.Tensor) -> float:
        return float(((x - y) ** 2).sum())

    @property
    def models(self) -> list[MLP]:
        return [MLP(5) for _ in range(3)]

    @property
    def train_dl(self) -> DataLoader[tuple[torch.Tensor, torch.Tensor]]:
        return DataLoader(RandomDataset((5,), 5), batch_size=1, shuffle=True)

    @property
    def val_dl(self) -> DataLoader[tuple[torch.Tensor, torch.Tensor]]:
        return DataLoader(RandomDataset((5,), 2), batch_size=1, shuffle=False)

    @property
    def eval_fn(self) -> Any:
        dl = self.val_dl

        def eval_fn(model: nn.Module, device: torch.device | str | None) -> float:
            loss = 0.0
            for x, y in dl:
                x, y = x.to(device), y.to(device)
                loss += self.loss_fn(model(x), y)
            return loss / len(dl)

        return eval_fn


class TestInterpolation(BaseClass):
    """Test sanity checks of Interpolation class."""

    def test_sanity_checks_model(self) -> None:
        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=[],
                eval_fn=self.eval_fn,
                eval_mode="min",
            )

        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=MLP(5),  # type: ignore[arg-type]
                eval_fn=self.eval_fn,
                eval_mode="min",
            )

        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=[*self.models, "not a model"],  # type: ignore[list-item]
                eval_fn=self.eval_fn,
                eval_mode="min",
            )

    def test_sanity_checks_eval_fn(self) -> None:
        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                eval_fn=0,
                eval_mode="min",
            )

    def test_sanity_checks_dataloader(self) -> None:
        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                eval_fn=self.eval_fn,
                eval_mode="min",
                train_dataloader=0,  # type: ignore[arg-type]
            )

    def test_sanity_checks_devices(self) -> None:
        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                eval_fn=self.eval_fn,
                eval_mode="min",
                devices=["cpu", "cpu", "cpu"],
                device_interp=None
            )

        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                eval_fn=self.eval_fn,
                eval_mode="min",
                devices=["cpu", "cpu", "cpu", "cpu"],  # too long
                device_interp="cpu"
            )

        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                eval_fn=self.eval_fn,
                eval_mode="min",
                devices=None,
                device_interp="cpu"
            )

    def test_sanity_checks_input_indices(self) -> None:
        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                eval_fn=self.eval_fn,
                eval_mode="min",
                input_indices=0.5  # type: ignore[arg-type]
            )

        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                eval_fn=self.eval_fn,
                eval_mode="min",
                input_indices=[0.5]  # type: ignore[list-item]
            )

        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                eval_fn=self.eval_fn,
                eval_mode="min",
                input_indices=["not", "an", "int"]  # type: ignore[list-item]
            )

    def test_sanity_checks_save_vars(self) -> None:
        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                eval_fn=self.eval_fn,
                eval_mode="min",
                savedir=0  # type: ignore[arg-type]
            )

        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                eval_fn=self.eval_fn,
                eval_mode="min",
                save_all="not a bool"  # type: ignore[arg-type]
            )

        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                eval_fn=self.eval_fn,
                eval_mode="min",
                save_all=True,
                savedir=None
            )

    def test_sanity_checks_verbose(self) -> None:
        with pytest.raises(AssertionError):
            interp.Interpolation(
                models=self.models,
                eval_fn=self.eval_fn,
                eval_mode="min",
                verbose="not a bool"  # type: ignore[arg-type]
            )


class TestLerpSimple(BaseClass):
    """Test the LerpSimple class."""
    @property
    def mlps(self) -> list[nn.Module]:
        return [MLP(5), MLP(5)]

    def test_settings_are_plausible(self) -> None:
        model_a, model_b = self.mlps

        # Must be different models
        for p1, p2 in zip(model_a.parameters(), model_b.parameters()):  # noqa
            assert not torch.allclose(p1, p2)

        # Must produce different outputs
        x = torch.rand(5)
        assert not torch.allclose(model_a(x), model_b(x))

        # Must have different losses
        val_dl = self.val_dl
        x, y = next(iter(val_dl))
        loss_a = loss_fn(model_a(x), y)
        loss_b = loss_fn(model_b(x), y)

        assert loss_a != loss_b

    def test_mlp(self) -> None:
        lerp = interp.LerpSimple(
            models=self.mlps,
            eval_fn=self.eval_fn,
            eval_mode="min",
        )
        lerp.interpolate(steps=10)

        assert len(lerp.metrics_interpolated) == 10
        assert len(lerp.metrics_models) == 2

        # With random data, there should be no duplicate losses.
        assert len(set(lerp.metrics_models)) == 2
        assert len(set(lerp.metrics_interpolated)) == 10

        # Check that the best model is the one with the lowest loss
        best_model_loss = min(lerp.metrics_models)
        best_interp_loss = min(lerp.metrics_interpolated)
        best_loss = min(best_model_loss, best_interp_loss)
        assert abs(best_loss - lerp.best_metric) < 1e-6 * lerp.best_metric


@pytest.mark.skipif(not torch.cuda.is_available(), reason="GPU tests")
class TestInterpolationGPU(BaseClass):
    def test_mlp(self) -> None:
        models = [MLP(5).to("cuda"), MLP(5)]
        devices = ["cuda", "cpu"]
        lerp = interp.LerpSimple(
            models=models,
            eval_fn=self.eval_fn,
            eval_mode="min",
            devices=devices,
            device_interp="cuda",
        )
        lerp.interpolate(steps=10)
        assert len(lerp.metrics_interpolated) == 10
        assert len(lerp.metrics_models) == 2

    @pytest.mark.skipif(torch.cuda.device_count() < 2, reason="Not enough GPUs")
    def test_mlp_multi_gpu(self) -> None:
        """Test whether it works with :code:`device_interp not in devices`."""

        models = [MLP(5).to("cuda:0"), MLP(5).to("cuda:1")]
        devices = ["cuda:0", "cuda:1"]
        lerp = interp.LerpSimple(
            models=models,
            eval_fn=self.eval_fn,
            eval_mode="min",
            devices=devices,
            device_interp="cuda",
        )
        lerp.interpolate(steps=10)
        assert len(lerp.metrics_interpolated) == 10
        assert len(lerp.metrics_models) == 2



from __future__ import annotations

import itertools
from collections.abc import Iterator, Sequence
from typing import Any

import torch
from torch import nn
from torch.utils.data import DataLoader
from tqdm import tqdm

from rebasin.structs import ModelInfo


def recalculate_batch_norms(
        model: nn.Module,
        dataloader: DataLoader[Any],
        input_indices: int | Sequence[int],
        device: torch.device | str | None,
        verbose: bool,
        *forward_args: Any,
        **forward_kwargs: Any
) -> None:
    """
    Recalculate the BatchNorm statistics of a model.

    Use this after permuting the model, but only if the model contains
    a BatchNorm.

    Args:
        model:
            The model.
        dataloader:
            The DataLoader to use for recalculating the statistics.
            Should ideally be the training DataLoader.
        input_indices:
            Many DataLoaders return several inputs and labels.
            These can sometimes be used in unexpected ways.
            To make sure that the correct outputs of the dataloader are used
            as inputs to the model's forward pass, you can specify the indices
            at which the inputs are located, in the order that they should be
            passed to the model.
        device:
            The device on which to run the model.
        verbose:
            Whether to print progress.
        *forward_args:
            Any additional positional arguments to pass to the model's forward  pass.
        **forward_kwargs:
            Any additional keyword arguments to pass to the model's forward pass.
    """
    if verbose:
        print("Recalculating BatchNorm statistics...")
    types = [type(m) for m in model.modules()]
    if (
            nn.BatchNorm1d not in types
            and nn.BatchNorm2d not in types
            and nn.BatchNorm3d not in types
    ):
        if verbose:
            print("No BatchNorm layers found in model.")
        return

    training = model.training
    model.train()

    # Reset the running mean and variance
    for module in model.modules():
        if (
                not isinstance(module, (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d))
                or not hasattr(module, "running_mean")
                or not hasattr(module, "running_var")
                or module.running_mean is None
                or module.running_var is None
        ):
            continue

        model.running_mean = torch.zeros_like(module.running_mean)
        model.running_var = torch.zeros_like(module.running_var)
    # Recalculate the running mean and variance
    for batch in tqdm(dataloader, disable=not verbose):
        if isinstance(batch, Sequence):
            inputs, _ = get_inputs_labels(batch, input_indices, 0, device)
        else:
            inputs = [batch]
        model(*inputs, *forward_args, **forward_kwargs)

    if not training:
        model.eval()


def get_inputs_labels(
        batch: Any,
        input_indices: int | Sequence[int] = 0,
        label_indices: int | Sequence[int] = 1,
        device: torch.device | str | int | None = None
) -> tuple[list[Any], list[Any]]:
    """
    Get the inputs and outputs from a batch.

    Args:
        batch:
            The batch.
        input_indices:
            Many DataLoaders return several inputs and labels per batch.
            These can sometimes be used in unexpected ways.
            To make sure that the correct outputs of the dataloader are used
            as inputs to the model's forward pass, you can specify the indices
            at which the inputs are located, in the order that they should be
            passed to the model.
        label_indices:
            Like `input_indices`, but for the labels.
        device:
            The device on which to run the model.

    Returns:
        The inputs and labels.
    """
    if isinstance(input_indices, int):
        input_indices = [input_indices]
    if isinstance(label_indices, int):
        label_indices = [label_indices]

    inputs = (
        [batch[i].to(device) for i in input_indices]
        if device is not None
        else [batch[i] for i in input_indices]
    )
    labels = (
        [batch[i].to(device) for i in label_indices]
        if device is not None
        else [batch[i] for i in label_indices]
    )
    return inputs, labels


def contains_parameter(
        parameters: Sequence[nn.Parameter] | Iterator[nn.Parameter],
        parameter: nn.Parameter
) -> bool:
    """
    Check if a sequence of parameters contains a parameter.

    This cannot be done via the normal `in` operator, because
    `nn.Parameter`'s `__eq__` does not work for parameters of different shapes.

    Args:
        parameters:
            The sequence of parameters.
        parameter:
            The parameter.

    Returns:
        Whether the sequence contains the parameter.
    """
    return any(param is parameter for param in parameters)


def model_info(model: nn.Module) -> ModelInfo:
    """
    Get information about a model.

    Args:
        model:
            The model.

    Returns:
        The :class:`ModelInfo` object containing the desired information.
    """
    name = model.__class__.__name__
    contains_batch_norm = any(
        isinstance(m, (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d))
        for m in model.modules()
    )
    num_parameters = sum(p.numel() for p in model.parameters())
    num_weights = sum(1 if "weight" in n else 0 for n, _ in model.named_parameters())
    num_biases = sum(1 if "bias" in n else 0 for n, _ in model.named_parameters())
    num_elements_per_weight = [
        p.numel() for n, p in model.named_parameters() if "weight" in n
    ]
    num_elements_per_bias = [
        p.numel() for n, p in model.named_parameters() if "bias" in n
    ]

    num_permutable_elements_per_weight = []
    for pname, parameter in model.named_parameters():
        if "weight" in pname:
            shape = parameter.shape
            permutable = shape[0] if len(shape) == 1 else shape[0] * shape[1]
            num_permutable_elements_per_weight.append(permutable)

    num_elements_per_weight_mean = torch.mean(
        torch.tensor(num_elements_per_weight, dtype=torch.float)
    ).item()
    num_elements_per_weight_std = torch.std(
        torch.tensor(num_elements_per_weight, dtype=torch.float)
    ).item()
    num_elements_per_bias_mean = torch.mean(
        torch.tensor(num_elements_per_bias, dtype=torch.float)
    ).item()
    num_elements_per_bias_std = torch.std(
        torch.tensor(num_elements_per_bias, dtype=torch.float)
    ).item()

    return ModelInfo(
        name=name,
        contains_batch_norm=contains_batch_norm,
        num_parameters=num_parameters,
        num_weights=num_weights,
        num_biases=num_biases,
        num_elements_per_weight=num_elements_per_weight,
        num_permutable_elements_per_weight=num_permutable_elements_per_weight,
        num_elements_per_bias=num_elements_per_bias,
        num_elements_per_weight_mean=num_elements_per_weight_mean,
        num_elements_per_weight_std=num_elements_per_weight_std,
        num_elements_per_bias_mean=num_elements_per_bias_mean,
        num_elements_per_bias_std=num_elements_per_bias_std,
    )


def pairwise(iterable: Sequence[Any]) -> Iterator[tuple[Any, Any]]:
    """
    Iterate over a sequence pairwise.

    Args:
        iterable:
            The sequence.

    Yields:
        The pairs.
    """
    try:
        return itertools.pairwise(iterable)
    except AttributeError:
        a, b = itertools.tee(iterable)
        next(b, None)
        return zip(a, b)  # noqa



"""Test the rebasin.initialization._paths module."""

from __future__ import annotations

import copy
from collections.abc import Callable

import torch
from torch import nn

from rebasin._paths import ModelPaths
from rebasin.structs import ModuleParameters
from tests.fixtures.paths import ModuleGenerator
from tests.fixtures.utils import allclose, model_change_percent


class TestPaths(ModuleGenerator):

    def test_linear_path(self) -> None:
        change_percs = torch.empty(10)

        for i in range(10):
            model, seq = self.linear_path()
            model_orig = copy.deepcopy(model)
            x = torch.randn(3, 3)
            y_orig = model(x)

            path = ModelPaths(seq)
            path.apply_permutations()

            # Check that the model has changed
            change_percs[i] = (model_change_percent(model, model_orig))

            # Check that the output nevertheless stays the same
            y_new = model(x)
            assert allclose(y_orig, y_new)

        assert change_percs.mean().item() > 0.1

    def test_dense_residual_path_simple(self) -> None:
        change_percs = torch.empty(10)

        for i in range(10):
            model, paths = self.residual_path_linear_simple()
            model_orig = copy.deepcopy(model)
            x = torch.randn(3, 3)
            y_orig = model(x)

            path = ModelPaths(paths)
            path.apply_permutations()

            # Check that the model has changed
            change_percs[i] = (model_change_percent(model, model_orig))

            # Check that the output nevertheless stays the same
            y_new = model(x)
            assert allclose(y_orig, y_new)

        assert change_percs.mean().item() > 0.1

    def test_dense_residual_path_complex(self) -> None:
        change_percs = torch.empty(10)

        for i in range(10):
            model, paths = self.residual_path_linear_complex()
            model_orig = copy.deepcopy(model)
            x = torch.randn(3, 3)
            y_orig = model(x)

            path = ModelPaths(paths)
            path.apply_permutations()

            # Check that the model has changed
            change_percs[i] = (model_change_percent(model, model_orig))

            # Check that the output nevertheless stays the same
            y_new = model(x)
            assert allclose(y_orig, y_new)

        assert change_percs.mean().item() > 0.1

    def test_combined_dense_paths(self) -> None:
        change_percs = torch.empty(10)

        for i in range(10):
            model_res_one, paths_res_one = self.residual_path_linear_complex()
            model_res_two, paths_res_two = self.residual_path_linear_simple()
            model_lin, paths_lin = self.linear_path()

            model = nn.Sequential(model_res_one, model_lin, model_res_two)
            model_orig = copy.deepcopy(model)
            x = torch.randn(3, 3)
            y_orig = model(x)

            path = ModelPaths(paths_res_one + paths_lin + paths_res_two)
            path.apply_permutations()

            # Check that the model has changed
            change_percs[i] = (model_change_percent(model, model_orig))

            # Check that the output nevertheless stays the same
            y_new = model(x)
            assert allclose(y_orig, y_new)

        assert change_percs.mean().item() > 0.1

    @staticmethod
    def path_test(
            func: Callable[
                [],
                tuple[torch.Tensor, nn.Module, list[list[ModuleParameters]]]
            ]
    ) -> None:
        change_percs = torch.empty(10)

        for i in range(10):
            x, model, paths = func()
            model_orig = copy.deepcopy(model)
            y_orig = model(x)

            path = ModelPaths(paths)
            path.apply_permutations()

            permstr = "\n".join(
                [
                    f"{p.module_type.__name__}:"
                    f"\n\tin: {p.input_permutation.perm_indices}"
                    f"\n\tout: {p.output_permutation.perm_indices}"
                    for p in path.paths[0]
                ]
            )
            print(f"{func.__name__}:\n" + permstr + "\n\n")

            # Check that the model has changed
            change_percs[i] = (model_change_percent(model, model_orig))

            # Check that the output nevertheless stays the same
            y_new = model(x)
            assert allclose(y_orig, y_new)

        assert change_percs.mean().item() > 0.1

    def test_linear_conv_path_simple(self) -> None:
        self.path_test(self.linear_path_conv2d_simple)

    def test_linear_conv_path_medium(self) -> None:
        self.path_test(self.linear_path_conv2d_medium)

    def test_linear_conv_path_complex(self) -> None:
        self.path_test(self.linear_path_conv2d_complex)

    def test_conv_residual_path_simple(self) -> None:
        self.path_test(self.residual_path_conv2d_simple)

    def test_linear_path_with_batchnorm(self) -> None:
        self.path_test(self.linear_path_with_batch_norm)
